#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <assert.h>
#include <unistd.h>
#include <errno.h>

#include <sys/stat.h>
#include <fcntl.h>

#include <ctype.h>
#include <string.h>

#include <arpa/inet.h>


#include "messages.h"
#include "log.h"

#define TRUE 1

#define ACCEPTING_PORT 4610
#define ACCEPT_QUEUE_LEN 10

static int validate_initial_request(struct initial_request *pireq) {
  /* checks the fields in the initial request */
  /* makes sure they've been encoded somewhat correctly */
  /* returns the number of the field that may be broken. */
  int i;
  i = ntohl(pireq->projectnumber);
  if(i > 3 || i <1) 
    return(2);
  /* studentid1 must be non-zero, and less than 10M */
  /* TODO: strengthen based on new stuids */
  i = ntohl(pireq->studentid1);
  if(i > 10000000  || i <=0) 
    return(3);
  /* students can't have started in 2001 or before 1990. */
  if(i > 100000  && i <= 9000000) 
    return(3);
  /* studentid2 may be zero */
  i = ntohl(pireq->studentid2);
  if(i > 10000000  || i <0) 
    return(4);
  if(i > 100000  && i <= 9000000) 
    return(4);
  /* name1 must have only printable characters */
  for(i=0; i<IREQ_MAX_NAME_LEN && pireq->name1[i] != '\0'; i++) {
    if(!isprint(pireq->name1[i]))
      return(5);
  }
  /* name1 must have something */
  if(i==0) return(5);
  /* name2 must have only printable characters, but may be empty */
  for(i=0; i<IREQ_MAX_NAME_LEN && pireq->name2[i] != '\0'; i++) {
    if(!isprint(pireq->name2[i]))
      return(6);
  }
  /* username1 must have only printable characters */
  for(i=0; i<IREQ_MAX_USERNAME_LEN && pireq->username1[i] != '\0'; i++) {
    if(!isprint(pireq->username1[i]))
      return(7);
  }

  /* username1 must have something */
  if(i==0) return(7);
  /* username2 must have only printable characters */
  for(i=0; i<IREQ_MAX_USERNAME_LEN && pireq->username2[i] != '\0'; i++) {
    if(!isprint(pireq->username2[i]))
      return(8);
  }
  for(i=0; i<25 && pireq->filename[i] != '\0'; i++) {
    if(!isprint(pireq->filename[i]))
      return(9);
  }
  return(0); /* got this far, didn't find an error */
}
static void write_metafile(const char *filename, 
                           struct initial_request *pireq) {
  FILE *f = fopen(filename, "w");
  if(f==NULL) {
    f = stdout; /* write the rest of the metafile to the screen */
    LOG("unable to fopen metafile\n");
  }
  fprintf(f, "Student ID 1: %d\n", ntohl(pireq->studentid1));
  fprintf(f, "Student ID 2: %d\n", ntohl(pireq->studentid2));
  fprintf(f, "Name 1: %s\n", pireq->name1);
  fprintf(f, "Name 2: %s\n", pireq->name2);
  fprintf(f, "username 1: %s\n", pireq->username1);
  fprintf(f, "username 2: %s\n", pireq->username2);
  fprintf(f, "filename: %s\n", pireq->filename);
  (void)fclose(f);
}

/* htonl and ntohl are duals - they do *exactly* the same
   thing, but I use ntohl when converting a network order
   long to a host order long for clarity */
static void handle_client(int s_client) {
  ssize_t bytes;
  struct initial_request ireq;
  struct initial_response iresp;
  struct upload_response uresp;
  unsigned int file_length, nfile_length;
  char filename[128];
  char metafilename[128];
  int r;
  FILE *turnin_file;
	int previousturnins;

  /* start the cleanup timer; ignore any previous alarms. */
  (void)alarm(5);
  /* read initial request */
  LOG("reading\n");
  bytes = read(s_client, &ireq, sizeof(struct initial_request));
  /* extra cruft in case the bytes come in multiple packets. */
  if(bytes ==  0) { 
    LOG("other side finished before it began\n");
    (void)close(s_client);
    return;
  }
  if(bytes < 0) {
    LOG2("other side errd %s\n", strerror(errno));
    (void)close(s_client);
    return;
  }
  LOG2("read %d byte(s)\n", bytes);
  while(bytes < 4 || bytes < ntohl(ireq.length)) {
    int readbytes;
    if(bytes >=4 && (ntohl(ireq.length) != sizeof(struct initial_request))) {
      /* unsupported or invalid initial request; bail early */
      LOG3("broken initial request length: %d (net: %d)\n",
				 	ntohl(ireq.length), ireq.length);
      iresp.error = htonl(1);
      goto BAIL;
    }
    readbytes = read(s_client, 
                     ((char *)&ireq) + bytes, 
                     sizeof(struct initial_request) - bytes);
    if(readbytes <= 0) {
      LOG("unexpected remote close.\n");
      (void)close(s_client);
    }
    bytes += readbytes;
  } 
  LOG("header read complete\n");
  /* end cruft */
  iresp.error = htonl(validate_initial_request(&ireq));
  /* write initial response */
 BAIL:
  LOG2("responding error %d\n", ntohl(iresp.error));
  iresp.length = htonl(sizeof(struct initial_response));
  (void)write(s_client, &iresp, sizeof(struct initial_response));
  if(ntohl(iresp.error) != 0) 
    return;
  /* read file */
  LOG("reading file length\n");
  bytes = 0;
  while(bytes < sizeof(int)) {
    ssize_t bytes_read = read(s_client, ((char *)&nfile_length)+bytes, sizeof(int)-bytes);
    bytes += bytes_read;
    if(bytes_read <= 0) {
      LOG("unexpected remote close on read\n");
      return; /* unexpected remote close */
    }
  }
  file_length = ntohl(nfile_length);

  LOG2("reading file of length %u\n", file_length);
  bytes = 0;
  previousturnins = 0;
  /* setup directories that should already be there. */
  mkdir("turnins", 0700);
  sprintf(filename, "turnins/%d", ntohl(ireq.projectnumber));
  mkdir(filename, 0700);
  do {
    sprintf(filename, "turnins/%d/%d.%d", 
          ntohl(ireq.projectnumber),
          ntohl(ireq.studentid1), 
          previousturnins);
    r = mkdir(filename, 0700) ;
    if(r!=0) 
      previousturnins++;
  } while(r!=0 && errno == EEXIST );

  strcat(filename, "/");
  strcpy(metafilename, filename);
  strcat(metafilename, "META");
  write_metafile(metafilename, &ireq);

  /* let's just not get confused. */
  assert(strcmp(filename, ireq.filename) != 0); 
  strncat(filename, ireq.filename, 25);
  
  turnin_file = fopen(filename, "w");
  assert(turnin_file != NULL);
  bytes = 0;
  while(bytes < file_length) {
    int readbytes;
    char buf[4096];
    readbytes = read(s_client, buf, 4096);
    LOG2("read %d\n", readbytes);
    if(readbytes <= 0) {
      LOG("unexpected remote close.\n");
      (void)close(s_client);
      return;
    }
    (void)fwrite(buf, readbytes, 1, turnin_file);
    bytes += readbytes;
  } 
  LOG("file read complete\n");
  (void)fclose(turnin_file);
  /* write ack */
  uresp.length = htonl(sizeof(struct upload_response));
  uresp.passed = htonl(1);
  uresp.percentage = htonl(100);
  /* store something. */
  (void)write(s_client, &uresp, sizeof(struct upload_response));
  (void)close(s_client);
  return;
}

int main(int argc, char *argv[]) {
  int s_accepting, s_client;
  struct sockaddr_in addr_client, addr_server;
  socklen_t len_addr_client = sizeof(struct sockaddr_in);
  int pid_child;
  int r;

  s_accepting=socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  assert(s_accepting>=0);
  
  memset(&addr_server, 0, sizeof(struct sockaddr_in));
  addr_server.sin_family = AF_INET;
  addr_server.sin_addr.s_addr = htonl(INADDR_ANY);
  addr_server.sin_port = htons(ACCEPTING_PORT);
  log_open();
  r = bind(s_accepting, 
           (struct sockaddr *) &addr_server, 
           sizeof(struct sockaddr_in));
	if(r!=0) {
		LOG("unable to bind\n");
	}
  assert(r==0);
  r = listen(s_accepting, ACCEPT_QUEUE_LEN);
  assert(r>=0);
  
  while(TRUE) {
    while(waitpid(-1, NULL, WNOHANG)>0); /* collect them all */
    s_client = accept(s_accepting, 
                      (struct sockaddr *) &addr_client, 
                      &len_addr_client);
    assert(s_client >= 0);
    pid_child = fork();
    assert( pid_child >= 0 );
    if(pid_child == 0) {
      /* the child */
      struct sockaddr_in remote;
      socklen_t len_remote = sizeof(struct sockaddr_in);
      (void)close(s_accepting);
      assert( getpeername(s_client, &remote, &len_remote) == 0);
      LOG2("connection from %s\n", inet_ntoa(remote.sin_addr));
#ifdef undef
      if(remote.sin_addr.s_addr != inet_addr("127.0.0.1") && /* loopback */
         remote.sin_addr.s_addr != inet_addr("128.95.8.16") && /* fiji */
         remote.sin_addr.s_addr != inet_addr("128.95.8.15") && /* ceylon */
         remote.sin_addr.s_addr != inet_addr("128.95.8.14") && /* sumatra */
         remote.sin_addr.s_addr != inet_addr("128.95.8.12")) { /* tahiti */
        LOG("not allowed.\n");
        (void)close(s_client);
        exit(EXIT_SUCCESS);
      }
#endif
      handle_client( s_client );
      exit(EXIT_SUCCESS);
    } else {
      /* the parent */
      (void)close(s_client);
    }
  }
}

      
    
  
  
  
